package com.s5gi.giGuns.guns;

import com.s5gi.giGuns.giGuns;
import org.bukkit.*;
import org.bukkit.block.Block;
import org.bukkit.entity.*;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.block.Action;
import org.bukkit.event.entity.ProjectileHitEvent;
import org.bukkit.event.player.PlayerInteractAtEntityEvent;
import org.bukkit.event.player.PlayerInteractEvent;
import org.bukkit.inventory.EquipmentSlot;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.PlayerInventory;
import org.bukkit.metadata.FixedMetadataValue;
import org.bukkit.util.Vector;

import java.util.*;

public class GunHandler implements Listener {
    public static ArrayList<IGun> guns = new ArrayList<>();
    public static HashMap<String, IGun> registeredGuns = new HashMap<>();
    public static HashMap<Snowball, IBullet> physicsBullets = new HashMap<>();
    public static HashMap<Snowball, Location> lastLocationsOfBullets = new HashMap<>();

    public static void register(IGun gun) {
        registeredGuns.put(giGuns.color(gun.name()), gun);
        guns.add(gun);
        giGuns.plugin.getServer().addRecipe(gun.recipe());
    }

    public static boolean isRegisteredGun(IGun gun) {
        return registeredGuns.containsValue(gun);
    }
    public static boolean isRegisteredGun(String name) {
        return registeredGuns.containsKey(giGuns.color(name));
    }
    public static IGun getRegisteredGun(String name) {
        return registeredGuns.get(giGuns.color(name));
    }

    public static void BulletRoutine() {
        List<Snowball> toRemove = new ArrayList<>();
        for (Snowball snowball : physicsBullets.keySet()) {
            if (System.currentTimeMillis() - snowball.getMetadata("bornAt").getFirst().asLong() >= 20*1000) {
                toRemove.add(snowball);
            }
        }
        for (Snowball snowball : toRemove) {
            snowball.remove();
            physicsBullets.remove(snowball);
        }
    }

    public static void BulletTrailRoutine() {
        for (Snowball physbullet : physicsBullets.keySet()) {
            Location currentLocation = physbullet.getLocation();
            Location lastLocation = lastLocationsOfBullets.get(physbullet);

            if (lastLocation != null) {
                Calc.drawParticleLine(lastLocation, currentLocation);
            }
            lastLocationsOfBullets.put(physbullet, currentLocation);
        }
    }

    public static void shoot(Player from) {
        World world = from.getWorld();
        IGun gun = GunHandler.getRegisteredGun(from.getInventory().getItemInMainHand().getItemMeta().getItemName());
        IBullet bullet = gun.compileBullet();
        if (from.hasCooldown(from.getInventory().getItemInMainHand())) return;
        if (!Calc.decreaseMatchingItem(from, gun.bulletType().getItem())) {
            world.playSound(from.getEyeLocation(), Sound.BLOCK_DISPENSER_FAIL, SoundCategory.PLAYERS, 3f, 1f);
            return;
        }
        Snowball physbullet = from.getWorld().spawn(from.getEyeLocation().subtract(0,0.2f,0), Snowball.class);
        physbullet.setShooter(from);
        physbullet.setItem(new ItemStack(bullet.mat()));
        physbullet.setGravity(gun.gravity());
        physbullet.setVelocity(from.getEyeLocation().getDirection().multiply(gun.bulletSpeed()));
        physbullet.setMetadata("isBullet", new FixedMetadataValue(giGuns.plugin, true));
        physbullet.setMetadata("bornAt", new FixedMetadataValue(giGuns.plugin, System.currentTimeMillis()));
        for (Sound sound : gun.sound()) {
            world.playSound(from.getEyeLocation(), sound, SoundCategory.PLAYERS, gun.soundVolume(), gun.soundPitch());
        }
        from.setCooldown(from.getInventory().getItemInMainHand(), gun.cooldown());
        lastLocationsOfBullets.put(physbullet, physbullet.getLocation());
        physicsBullets.put(physbullet, bullet);
    }

    @EventHandler
    public static void onRightClick(PlayerInteractEvent event) {
        if (event.getAction() == Action.RIGHT_CLICK_BLOCK || event.getAction() == Action.RIGHT_CLICK_AIR) {
            if (event.getAction()==Action.RIGHT_CLICK_BLOCK && Calc.isInteractable(event.getClickedBlock()) && !event.getPlayer().isSneaking()) {
                return;
            }
            if (event.getHand() == EquipmentSlot.HAND) {
                if (event.getPlayer().getInventory().getItemInMainHand() !=null && event.getPlayer().getInventory().getItemInMainHand().getItemMeta()!=null  && GunHandler.isRegisteredGun(event.getPlayer().getInventory().getItemInMainHand().getItemMeta().getItemName())) {
                    event.setCancelled(true);
                    shoot(event.getPlayer());
                } else if (event.getPlayer().getInventory().getItemInMainHand() != null &&
                        event.getPlayer().getInventory().getItemInMainHand().getType().equals(Material.BLACK_DYE) &&
                        event.getPlayer().getInventory().getItemInMainHand().getItemMeta()!=null &&
                        event.getPlayer().getInventory().getItemInMainHand().getItemMeta().getItemName().contains("Ammo")) {
                    event.setCancelled(true);
                }
            } else if (event.getHand() == EquipmentSlot.OFF_HAND) {
                if (event.getPlayer().getInventory().getItemInOffHand() !=null && event.getPlayer().getInventory().getItemInOffHand().getItemMeta()!=null  && GunHandler.isRegisteredGun(event.getPlayer().getInventory().getItemInOffHand().getItemMeta().getItemName())) {
                    event.setCancelled(true);
                }
                else if (event.getPlayer().getInventory().getItemInOffHand() != null &&
                        event.getPlayer().getInventory().getItemInOffHand().getType().equals(Material.BLACK_DYE) &&
                        event.getPlayer().getInventory().getItemInOffHand().getItemMeta()!=null &&
                        event.getPlayer().getInventory().getItemInOffHand().getItemMeta().getItemName().contains("Ammo")) {
                    event.setCancelled(true);
                }

            }
        }
    }
    @EventHandler
    public static void onPlayerInteract(PlayerInteractAtEntityEvent event) {
        if (event.getPlayer().getInventory().getItemInMainHand() !=null && event.getPlayer().getInventory().getItemInMainHand().getItemMeta()!=null &&
                GunHandler.isRegisteredGun(event.getPlayer().getInventory().getItemInMainHand().getItemMeta().getItemName())) {
            if (event.getRightClicked() instanceof Villager || event.getRightClicked() instanceof ItemFrame) {return;}
            event.setCancelled(true);
            shoot(event.getPlayer());
        } else if (event.getPlayer().getInventory().getItemInMainHand() != null &&
                event.getPlayer().getInventory().getItemInMainHand().getType().equals(Material.BLACK_DYE) &&
                event.getPlayer().getInventory().getItemInMainHand().getItemMeta()!=null &&
                event.getPlayer().getInventory().getItemInMainHand().getItemMeta().getItemName().contains("Ammo")) {
            event.setCancelled(true);
        }
    }

    @EventHandler
    public static void onBulletHit(ProjectileHitEvent event) {
        if (event.getEntity() instanceof Snowball physbullet && physicsBullets.containsKey(physbullet)) {
            Player bulletOwner = (Player) event.getEntity().getShooter();
            IBullet bullet = physicsBullets.get(physbullet);
            IGun gun = bullet.parentGun();
            if (event.getHitEntity() instanceof Player damagedPlayer) {
                if (bulletOwner == damagedPlayer) {
                    event.setCancelled(true);
                    return;
                } else {
                    double dueDamage = gun.damage()-GunHandler.Calc.getDamageAfterArmor(damagedPlayer);
                    if (dueDamage < 0.5) dueDamage=0.5;
                    damagedPlayer.damage(dueDamage, bulletOwner);
                    gun.onHitEntity(damagedPlayer.getLocation(), damagedPlayer);
                    if (gun.makeBlood()) Calc.makeBlood(event.getEntity().getLocation(), (float) dueDamage);
                    physicsBullets.remove(physbullet);
                }
            } else if (event.getHitEntity() instanceof LivingEntity livingEntity) {
                livingEntity.damage(gun.damage(), bulletOwner);
                gun.onHitEntity(event.getEntity().getLocation(), livingEntity);
                if (gun.makeBlood()) Calc.makeBlood(event.getEntity().getLocation(), (int) gun.damage());
                physicsBullets.remove(physbullet);

            }
            if (event.getHitBlock()!=null) {
                event.getHitBlock().getDrops().clear();
                gun.onHitBlock(event.getEntity().getLocation(), event.getHitBlock());
                physicsBullets.remove(physbullet);
            }

        }
    }

    public static class Calc {
        public static double getDamageAfterArmor(Player player)
        {
            PlayerInventory inv = player.getInventory();

            Material boots = Material.AIR;
            Material helmet = Material.AIR;
            Material chest = Material.AIR;
            Material pants = Material.AIR;
            if (inv.getBoots()!=null) boots = inv.getBoots().getType();
            if (inv.getHelmet()!=null) boots = inv.getHelmet().getType();
            if (inv.getChestplate()!=null) boots = inv.getChestplate().getType();
            if (inv.getLeggings()!=null) boots = inv.getLeggings().getType();



            double red = 0.0;
            if(helmet == Material.LEATHER_HELMET)red = red + 0.04;
            else if(helmet == Material.GOLDEN_HELMET)red = red + 0.08;
            else if(helmet == Material.CHAINMAIL_HELMET)red = red + 0.08;
            else if(helmet == Material.IRON_HELMET)red = red + 0.08;
            else if(helmet == Material.DIAMOND_HELMET)red = red + 0.12;
            else if(helmet == Material.NETHERITE_HELMET)red = red + 0.16;
            //
            if(boots == Material.LEATHER_BOOTS)red = red + 0.04;
            else if(boots == Material.GOLDEN_BOOTS)red = red + 0.04;
            else if(boots == Material.CHAINMAIL_BOOTS)red = red + 0.04;
            else if(boots == Material.IRON_BOOTS)red = red + 0.08;
            else if(boots == Material.DIAMOND_BOOTS)red = red + 0.12;
            else if(boots == Material.NETHERITE_BOOTS)red = red + 0.16;

            //
            if(pants == Material.LEATHER_LEGGINGS)red = red + 0.08;
            else if(pants == Material.GOLDEN_LEGGINGS)red = red + 0.12;
            else if(pants == Material.CHAINMAIL_LEGGINGS)red = red + 0.16;
            else if(pants == Material.IRON_LEGGINGS)red = red + 0.20;
            else if(pants == Material.DIAMOND_LEGGINGS)red = red + 0.24;
            else if(pants == Material.NETHERITE_LEGGINGS)red = red + 0.28;

            //
            if(chest == Material.LEATHER_CHESTPLATE)red = red + 0.12;
            else if(chest == Material.GOLDEN_CHESTPLATE)red = red + 0.20;
            else if(chest == Material.CHAINMAIL_CHESTPLATE)red = red + 0.20;
            else if(chest == Material.IRON_CHESTPLATE)red = red + 0.24;
            else if(chest == Material.DIAMOND_CHESTPLATE)red = red + 0.32;
            else if(chest == Material.NETHERITE_CHESTPLATE)red = red + 0.38;

            return red * 2;
        }

        public static boolean decreaseMatchingItem(Player player, ItemStack targetItem) {
            PlayerInventory inventory = player.getInventory();
            if (!player.getGameMode().equals(GameMode.CREATIVE)) {
                for (int i = 0; i < inventory.getSize(); i++) {
                    ItemStack item = inventory.getItem(i);
                    if (item != null && item.isSimilar(targetItem)) {
                        if (item.getAmount() > 1) {
                            item.setAmount(item.getAmount() - 1);  // Decrease stack size
                        } else {
                            inventory.setItem(i, null);  // Remove item if it was the last one
                        }
                        return true;  // Found and processed the item
                    }
                }
            }
            return player.getGameMode().equals(GameMode.CREATIVE);  // No matching item found
        }

        public static void drawParticleLine(Location from, Location to) {
            Vector direction = to.toVector().subtract(from.toVector());
            double distance = direction.length();
            direction.normalize();

            for (double d = 0; d < distance; d += 0.5) {
                Location point = from.clone().add(direction.clone().multiply(d));
                from.getWorld().spawnParticle(Particle.DUST, point, 3, 0.01f, 0.01f, 0.01f, 0, new Particle.DustOptions(Color.WHITE, 0.5f));

            }
        }

        public static void makeBlood(Location location, float damage) {
            location.getWorld().spawnParticle(Particle.BLOCK_CRUMBLE, location, (int) damage * 8, 0.1f, 0.1f, 0.1f, Material.RED_WOOL.createBlockData());
        }

        private static final Set<Material> INTERACTABLE_BLOCKS = EnumSet.of(
                Material.CHEST,
                Material.TRAPPED_CHEST,
                Material.FURNACE,
                Material.BLAST_FURNACE,
                Material.SMOKER,
                Material.CRAFTING_TABLE,
                Material.ANVIL,
                Material.CHIPPED_ANVIL,
                Material.DAMAGED_ANVIL,
                Material.ENCHANTING_TABLE,
                Material.ENDER_CHEST,
                Material.BARREL,
                Material.LECTERN,
                Material.STONECUTTER,
                Material.GRINDSTONE,
                Material.CARTOGRAPHY_TABLE,
                Material.LOOM,
                Material.SMITHING_TABLE,
                Material.BEACON,
                Material.JUKEBOX,
                Material.NOTE_BLOCK,
                Material.BELL,
                Material.LEVER,
                Material.STONE_BUTTON,
                Material.OAK_BUTTON,
                Material.SPRUCE_BUTTON,
                Material.BIRCH_BUTTON,
                Material.JUNGLE_BUTTON,
                Material.ACACIA_BUTTON,
                Material.DARK_OAK_BUTTON,
                Material.CRIMSON_BUTTON,
                Material.WARPED_BUTTON,
                Material.POLISHED_BLACKSTONE_BUTTON,
                Material.OAK_DOOR,
                Material.SPRUCE_DOOR,
                Material.BIRCH_DOOR,
                Material.JUNGLE_DOOR,
                Material.ACACIA_DOOR,
                Material.DARK_OAK_DOOR,
                Material.CRIMSON_DOOR,
                Material.WARPED_DOOR,
                Material.IRON_DOOR,
                Material.OAK_TRAPDOOR,
                Material.SPRUCE_TRAPDOOR,
                Material.BIRCH_TRAPDOOR,
                Material.JUNGLE_TRAPDOOR,
                Material.ACACIA_TRAPDOOR,
                Material.DARK_OAK_TRAPDOOR,
                Material.CRIMSON_TRAPDOOR,
                Material.WARPED_TRAPDOOR,
                Material.IRON_TRAPDOOR,
                Material.OAK_FENCE_GATE,
                Material.SPRUCE_FENCE_GATE,
                Material.BIRCH_FENCE_GATE,
                Material.JUNGLE_FENCE_GATE,
                Material.ACACIA_FENCE_GATE,
                Material.DARK_OAK_FENCE_GATE,
                Material.CRIMSON_FENCE_GATE,
                Material.WARPED_FENCE_GATE
        );

        public static boolean isInteractable(Block block) {
            return INTERACTABLE_BLOCKS.contains(block.getType());
        }
    }
}
